import pyvisa
from ..utils import floatify, intify

class KeithleyCS:
    def __init__(self, rm: pyvisa.ResourceManager, dev='KEITHLEY', **ignored):
        self._res = rm.open_resource(dev, write_termination='\n', read_termination='\n')
        self._res.write("*RST")

    @property
    def current():
        raise NotImplementedError()
    @current.setter
    def current(self, v):
        self._res.write("SOUR:CURR " + floatify(v))

    def set_sweep(self, currents:list, trig_line:int = 1, count:int=1):
        if len(currents) == 0:
            raise ValueError(f"len(currents) should be non-zero")
        if trig_line not in range(1, 6+1):
            raise ValueError(f"trig_line ({trig_line}) is not an integer from 1 to 6")
        if not isinstance(count, int) or count < 1:
            raise ValueError(f"count ({count}) should be a positive integer")
        self._res.write("SOUR:SWE:SPAC LIST")
        self._res.write("SOUR:SWE:RANG BEST")
        self._res.write("SOUR:LIST:CURR " + ','.join([floatify(c, 9) for c in currents]))
        self._res.write("SOUR:LIST:DEL " + ','.join(['1e-3'] * len(currents)))
        self._res.write("SOUR:LIST:COMP " + ','.join(['10'] * len(currents)))
        self._res.write("SOUR:SWE:COUN " + intify(count))
        self._res.write("SOUR:SWE:CAB OFF")
        self._res.write("TRIG:SOUR TLIN")
        self._res.write("TRIG:ILIN " + intify(trig_line))
        self._res.write("SOUR:SWE:ARM")

    def init(self):
        # self._res.write("SOUR:SWE:ARM")
        self._res.write("INIT")

    def set_remote_only(self, b:bool=True):
        s = "OFF" if b else "ON"
        self._res.write("DISP:ENAB " + s)

    def on(self):
        self._res.write("OUTP ON")

    def off(self):
        self._res.write("OUTP OFF")




class ScpiParser:

    def __init__(self):
        self._commands = {}
        self._aliases = {}

    def register(self, cmd_form, fun):
        cmd = cmd_form.split(':')        
        for long_form in cmd:
            short_form =  ''.join(c for c in long_form if not c.islower())
            short_lc, long_lc = short_form.lower(), long_form.lower()
            if short_lc != long_lc and any(e.isupper() for e in short_form):
                self._aliases[short_lc] = long_lc
        cmd_expand = cmd_form.lower()
        self._commands[cmd_expand] = fun

    def _execute_task(self, task):
        cmd, args = task
        if "" in cmd or "" in args:
            # print("Improper command: empty values in", task)
            return
        cmd_lc = [c.lower() for c in cmd]
        cmd_full = [self._aliases.get(c, c) for c in cmd_lc]
        cmd_expand = ":".join(cmd_full)
        if cmd_expand in self._commands:
            return self._commands[cmd_expand](args)

    def _parse(self, data):
        cmd_tree = []
        tasks = []
        tasks_raw = [e for e in data.strip().split(';') if e.strip()]
        for task in tasks_raw:
            cmd_raw, args_raw = (task.strip() + ' ').split(' ', 1)
            cmd = cmd_raw.split(':')
            if cmd[0] and cmd[0][0] == "*":
                pass
            elif cmd[0]:
                cmd = cmd_tree + cmd
            else:
                cmd = cmd[1:]
            cmd_tree = cmd[:-1]
            args = []
            args_raw = args_raw.strip(",")
            if args_raw:
                args = [arg.strip() for arg in args_raw.split(',')]
            tasks.append([cmd, args])
        return tasks

    def _stringify(self, repl):
        if isinstance(repl, list) or isinstance(repl, tuple):
            return ','.join([str(v) for v in repl])
        else:
            return str(repl)            
        
    def process(self, data):
        tasks = self._parse(data)
        repls = []
        for task in tasks:
            repl = self._execute_task(task)
            if repl:
                repls.append(self._stringify(repl))
        # print(tasks)
        return ';'.join(repls)


def _echo(args):
    return ','.join(args)

if(__name__ == "__main__"):
    parser = ScpiParser()
    parser.register("MEASure?", _echo)
    parser.register("MEASure", _echo)
    print(parser.process("measure:stuff volt; *RST 1; :yolo"))


from .series import Series
import numpy as np
from scipy import signal

def fft(ser, pad = 1):
        real = np.isrealobj(ser._y)
        ft = np.fft.rfft if real else np.fft.fft
        ftfreq = np.fft.rfftfreq if real else np.fft.fftfreq
        d = abs(ser._x[1] - ser._x[0])
        if pad < 1.:
            raise ValueError(f"Padding should be >= 1, is {pad}")
        n = int(ser._x.size * pad)
        return Series(ft(ser._y, n=n), ftfreq(n, d))

def filter(ser, ker):
        return Series(signal.convolve(ser._y, ker, mode='same'), ser._x)

def project(ser, t0, lag=None, forward=False, taps=None, trend='c', info={}):
    try:
        from statsmodels.tsa.ar_model import AutoReg
        from statsmodels.tsa.ar_model import ar_select_order
    except ImportError as e:
        print(e)
        print("Install statsmodels manually or install labpython with [full] option e.g. pip install labpython[full]")
    ret = ser.copy_y()
    p1, p2 = ret.split(t0)
    if forward:
        train = p1.y
        to_pred = p2.y
    else:
        train = p2.y[::-1]
        to_pred = p1.y[::-1]
    if lag is None and taps is None:
        lag = ar_select_order(train, maxlag=50, trend=trend).ar_lags
    elif taps is not None:
        lag = taps
    else:
        lag = int(lag * ser.freq)
    info['lag'] = lag
    fit = AutoReg(train, lags=lag, trend=trend).fit()
    info['params'] = fit.params
    to_pred[:] = fit.model.predict(fit.params, start=train.size, end=train.size + to_pred.size -1)
    return ret



import pyvisa
from enum import Enum
from ..utils import intify, floatify, to_enum

class Srs:

    class Trigger(Enum):
        SINE_ZERO = 0
        RISING = 1
        FALLING = 2

    class Source(Enum):
        EXTERNAL = 0
        INTERNAL = 1

    class Reserve(Enum):
        HIGH_RESERVE = 0
        NORMAL = 1
        LOW_NOISE = 2

    class Input(Enum):
        X, Y, R, TH = 1, 2, 3, 4
        AUX1, AUX2, AUX3, AUX4 = 5, 6, 7, 8
        FREQ, DISP1, DISP2 = 9, 10, 11

    Sensitivity = {
        "2 nV": 0, "5 nV": 1, "10 nV": 2, "20 nV": 3, "50 nV": 4, "100 nV": 5,
        "200 nV": 6, "500 nV": 7, "1 uV": 8, "2 uV": 9, "5 uV": 10, "10 uV": 11,
        "20 uV": 12, "50 uV": 13, "100 uV": 14, "200 uV": 15, "500 uV": 16, "1 mV": 17,
        "2 mV": 18, "5 mV": 19, "10 mV": 20, "20 mV": 21,"50 mV": 22, "100 mV": 23,
        "200 mV": 24, "500 mV": 25, "1 V": 26,
    }
    SensitivityInv = {v: k for k, v in Sensitivity.items()}

    TimeConstant = {
        "10 us": 0, "30 us": 1, "100 us": 2, "300 us": 3, "1 ms": 4, "3 ms": 5,
        "10 ms": 6, "30 ms": 7, "100 ms": 8, "300 ms": 9, "1 s": 10, "3 s": 11,
        "10 s": 12, "30 s": 13, "100 s": 14, "300 s": 15, "1 ks": 16, "3 ks": 17,
        "10 ks": 18, "30 ks": 19,
    }
    TimeConstantInv = {v: k for k, v in TimeConstant.items()}

    FilterSlope = {
        "6 dB/oct": 0, "12 dB/oct": 1, "18 dB/oct": 2, "24 dB/oct": 3,
    }
    FilterSlopeInv = {v: k for k, v in FilterSlope.items()}


    def __init__(self, rm: pyvisa.ResourceManager, dev='Lock-in',
                 auxout_map = {}, auxin_map = {}, settings={}, **ignored):
        self._res = rm.open_resource(dev, write_termination='\n', read_termination='\n')
        self.auxout_map = auxout_map
        self.auxin_map = auxin_map
        self.setup(settings)

    def setup(self, attrs: dict):
        for key, value in attrs.items():
            setattr(self, key, value)

    @property
    def identity(self):
        return self._res.query("*IDN?")

    @property
    def source(self):
        return self.Source(int(self._res.query("FMOD?"))).name
    @source.setter
    def source(self, v: Source):
        self._res.write("FMOD " + str(to_enum(v, self.Source).value))

    @property
    def reserve(self):
        return self.Reserve(int(self._res.query("RMOD?"))).name
    @reserve.setter
    def reserve(self, v: Reserve):
        self._res.write("RMOD " + str(to_enum(v, self.Reserve).value))

    @property
    def frequency(self):
        """Some property getter"""
        return float(self._res.query("FREQ?"))
    @frequency.setter
    def frequency(self, v):
        self._res.write("FREQ " + floatify(v))

    @property
    def phase(self):
        return float(self._res.query("PHAS?"))
    @phase.setter
    def phase(self, v):
        if not -360. <= v < 730:
            raise ValueError(f"Phase value {v} is not in range [-360, 730)")
        self._res.write("PHAS " + floatify(v))

    @property
    def harmonic(self):
        return int(self._res.query("HARM?"))
    @harmonic.setter
    def harmonic(self, v):
        if not int(v) in range(1, 20000):
            raise ValueError(f"Harmonic value {v} is not in range [1, 19999]")
        self._res.write("HARM " + intify(v))

    @property
    def sensitivity(self):
        return self.SensitivityInv[int(self._res.query("SENS?"))]
    @sensitivity.setter
    def sensitivity(self, v):
        self._res.write("SENS " + str(self.Sensitivity[v]))

    @property
    def time_constant(self):
        return self.TimeConstantInv[int(self._res.query("OFLT?"))]
    @time_constant.setter
    def time_constant(self, v):
        self._res.write("OFLT " + str(self.TimeConstant[v]))

    @property
    def filter_slope(self):
        return self.FilterSlopeInv[int(self._res.query("OFSL?"))]
    @filter_slope.setter
    def filter_slope(self, v):
        self._res.write("OFSL " + str(self.FilterSlope[v]))

    def snap(self, params):
        if not len(params) in range (2, 6+1):
            raise ValueError(f"Min. 2, max. 6 values can be snapped at once (not {len(params)})")
        params_str = [str(to_enum(p, self.Input).value) for p in params]
        res = self._res.query("SNAP? " + ','.join(params_str))
        return [float(v) for v in res.split(',')]

    def demod(self, param):
        e = to_enum(param, self.Input)
        if e.value not in range(1, 4+1):
            raise ValueError(f"{e.name} is not a demodulated channel")
        return self._res.query("OUTP? " + str(e.value))

    def auxin(self, num):
        if isinstance(num, str):
            num = self.auxin_map[num]
        if not isinstance(num, int) or not num in range(1, 4+1):
            raise ValueError(f"{num} is not an integer from 1 to 4")
        return self._res.query("OAUX? " + str(num))

    def auxout(self, num, value: float = None):
        if isinstance(num, str):
            num = self.auxout_map[num]     
        if not isinstance(num, int) or not num in range(1, 4+1):
            raise ValueError(f"{num} is not an integer from 1 to 4")
        if value is None:
            return self._res.query("AUXV? " + str(num))
        else:
            self._res.write("AUXV " + str(num) + ',' + floatify(value, 3))

    @property
    def x(self):
        return self.demod(self.Input.X)
    @property
    def y(self):
        return self.demod(self.Input.Y)
    @property
    def r(self):
        return self.demod(self.Input.R)
    @property
    def th(self):
        return self.demod(self.Input.TH)
    @property
    def xy(self):
        return self.snap('x','y')
    @property
    def rth(self):
        return self.snap(('r','th'))

if __name__ == "__main__":

    visa = pyvisa.ResourceManager()
    lockin = Srs(visa)
    print(lockin.identity)
    lockin.source = Srs.Source.INTERNAL
    lockin.reserve = Srs.Reserve.NORMAL
    lockin.frequency = 500.16
    setattr(lockin, "frequency", 512.)
    lockin.phase = 0
    lockin.harmonic = 1
    lockin.sensitivity = "5 mV"
    lockin.time_constant = "1 ms"
    lockin.filter_slope = "12 dB/oct"
    print(lockin.source)
    print(lockin.reserve)
    print(lockin.frequency)
    print(lockin.phase)
    print(lockin.harmonic)
    print(lockin.sensitivity)
    print(lockin.time_constant)
    print(lockin.filter_slope)
    print(lockin.demod('x'))
    print(lockin.x)
    print(lockin.snap('x', 'y'))
    In = Srs.Input
    print(lockin.snap(In.AUX3, In.X, In.Y))
    print(lockin.snap(('x', 'y')))
    print(lockin.aux(1))
    print(lockin.xy)
    print(lockin.rth)



import numpy as np
from typing import Union
from math import inf, isclose, ceil

def _check_type(types, *vars):
    for v in vars:
        if not isinstance(v, types):
            raise TypeError(f"{type(v)} is not of type(s) {types}")

class Series:
    '''Series object represents ordered one-dimensional data.\n
    `y` property stores data (of any type accepted by `np.ndarray`), while `x` property stores equidistant, ascending
    ordering values (e.g. time or frequency).
    It supports `+`, `-`, `*`, `/`, `+=`, `-=`, `*=`, `/=` operators with some restrictions:
    - Series object must be left-hand side (lhs)
    - On right-hand side (rhs) either another `Series` object can be used
    or any type that is supported by underlying numpy array (eg. another numpy array or a single number)
    - If operation is performed on two `Series` objects, their x arrays must match,
    if numpy array is on rhs, its size must match with lhs `Series` object y array size

    Notes
    -----
    When new `Series` object is constructed, either explicitly with `Series()`
    or implicitly, e.g. when performing arithmetical operations or using functions
    like `slice`, `split`, only array references are copied if possible.
    Thus, if you modify x or y array in-place, changes may affect other objects.
    If you want to modify underlying arrays in-place, it's advised to make an explicit copy,
    either of both arrays with `copy` method or only of y array with `copy_y`

    See also
    --------
    `labpy.dsp`:
        collection of functions for processing data stored in `Series` object
    '''

    def __init__(self, y: np.ndarray, x: Union[np.ndarray, float] = None, freq: float = None, x0: float = 0.):
        '''Initialize Series instance. Numpy arrays passed are shallow-copied.

        Parameters
        ----------
        y: np.ndarray | Series-like
            Array with values or Series-like (with `x` and `y` attributes) object.
        x: np.ndarray | float
            Array of equidistant ordering values (e.g. time or frequency)
            or single value equal to length of measurement (e.g. total time). Exclusive with `freq` parameter.
        freq: float
            Frequency of samples. Exclusive with `x` parameter.
        x0: float
            Value added to x array when it's not passed explicitly,
            i.e. is defined by `freq` or single-valued `x` parameter. Ignored otherwise.
        '''
        if x is None and freq is None:
            # Assume x is Series-like and try to copy attributes
            x = np.asarray(y.x)
            y = np.asarray(y.y)
        self._x: np.ndarray = Series.calc_x(y, x, freq, x0)
        self._y: np.ndarray = np.asarray(y)

    def copy(self):
        return Series(self._y.copy(), self._x.copy())

    def copy_y(self):
        return Series(self._y.copy(), self._x)

    def decimate(self, samples: int = None, freq: float = None):
        if samples is not None and freq is not None:
            raise ValueError("Either samples or freq should be specified, not both")
        if freq is not None:
            samples = max(1, int(self.freq // freq))
        if samples is not None:
            return Series(self._y[::samples], self._x[::samples])
        else:
            raise ValueError("Either samples or freq should be specified")

    @property
    def xy(self):
        return (self._x, self._y)

    @property
    def x(self):
        return self._x
    @x.setter
    def x(self, x: np.ndarray):
        _check_type(np.ndarray, x)
        if(x.shape != self._y.shape):
            raise ValueError(f"Array x (shape={x.shape}) should be of the same shape as y (shape={self._y.shape})")
        self._x = x

    @property
    def y(self):
        return self._y
    @y.setter
    def y(self, y: np.ndarray):
        _check_type(np.ndarray, y)
        if(y.shape != self._x.shape):
            raise ValueError(f"Array y (shape={y.shape}) should be of the same shape as x (shape={self._x.shape})")
        self._y = y

    @property
    def range(self):
        return self._x[0], self._x[-1], self._x.size
    
    @property
    def dx(self):
        return abs(self._x[1] - self._x[0])

    @property
    def freq(self):
        return 1./abs(self._x[1] - self._x[0])

    def apply(self, fun):
        return Series(fun(self._x), self._x)

    def slice(self, l=-inf, r=inf, rel=False):
        '''Return new `Series` object constructed by taking view of `x` array
        for values $l \\leq x < r$ and corrresponding view of `y` array.
        If `rel = False` (default), $l$ and $r$ are absolute.
        If `rel = True`, $l$ is relative to `x` array beginning if $l \\geq 0$ and end if $l < 0$,
        same for $r$ except it points to `x` array end if $r = 0$\n
        Use `cut` for default relative specification.
        '''
        rng = self.range
        if rel:
            l = rng[0] + l if l >= 0. else rng[1] + l + self.dx
            r = rng[1] + self.dx + r if r <= 0. else rng[0] + r
        i0 = Series.find_idx(l, rng)
        i1 = Series.find_idx(r, rng)
        return Series(self._y[i0:i1], self._x[i0:i1])

    def cut(self, l=-inf, r=inf):
        '''Same as `slice` with `rel = True`'''
        return self.slice(l, r, rel=True)

    def extend(self, l=None, r=None, rel=False):
        rng = self.range
        s = rng[2]
        l0, r0 = rng[0], rng[1] + self.dx
        if l is not None:
            if rel: l = l0 + l
            li = min(Series.find_idx(l, rng, norm=False), 0)
            lx = np.linspace(l0 + li * (r0 - l0) / s, l0, -li, endpoint=False)
        else: lx = np.array([])
        if r is not None:
            if rel: r = r0 + r
            ri = max(Series.find_idx(r, rng, norm=False), s)
            rx = np.linspace(r0, r0 + (ri - s) * (r0 - l0) / s, (ri - s), endpoint=False)
        else: rx = np.array([])
        return Series(
            np.concatenate([np.zeros_like(lx), self._y, np.zeros_like(rx)]),
            np.concatenate([lx, self._x, rx])
        )

    def split(self, s, rel=False):        
        rng = self.range
        if rel:
            s = rng[0] + s if s >= 0. else rng[1] + s + self.dx
        i = Series.find_idx(s, rng)
        x, y = self._x, self._y
        return Series(y[:i], x[:i]), Series(y[i:], x[i:])

    def part(self, beg=0., end=1.):
        s = self._y.size
        r = (0., 1. , s + 1)
        i0 = min(Series.find_idx(beg, r), s)
        i1 = min(Series.find_idx(end, r), s)
        return Series(self._y[i0:i1], self._x[i0:i1])

    def __repr__(self):
        return f"D = {self.range} y = {self._y}"

    def _op_helper(self, other, op):
        if isinstance(other, Series):
            if(self.range != other.range):
                raise ValueError(f"Series' ranges should be equal, are {self.range} and {other.range}")
            other = other.y
        if op.startswith('__i'):
            self.y = getattr(self._y, op)(other)
            return self
        else:
            return Series(getattr(self._y, op)(other), self._x)

    @staticmethod
    def find_idx(v, range, norm=True):
        l, r, s = range
        if norm:
            if v <= l or isclose(v, l):
                return 0
            if v > r and not isclose(v, r):
                return s
        i: float = (v - l)/(r - l) * (s - 1)
        if isclose(i, round(i)):
            return int(i)
        return ceil(i)

    @staticmethod
    def from2darray(y2d: np.ndarray, *args, **kwargs):
        if(y2d.ndim != 2):
            raise ValueError("y2d array should be two-dimensional")
        nrow = y2d.shape[0]
        x = Series.calc_x(y2d[0], *args, **kwargs)
        return [Series(y2d[i], x) for i in range(0, nrow)]

    @staticmethod
    def calc_x(y: np.ndarray, x: Union[np.ndarray, float] = None, freq=None, x0 = 0.):
        _check_type(np.ndarray, y)
        if y.ndim != 1:
            raise ValueError(f"Array y (dim={y.ndim}), should be one-dimensional")
        if x is not None and freq is not None:
            raise ValueError("Either x or freq should be specified, not both")
        if x is None and freq is None:
            raise ValueError("Either x or freq should be specified")
        if freq is not None:
            x = y.size / freq
        if not isinstance(x, np.ndarray):
            l, r = x0, x0 + x
            x = np.linspace(l, r, y.size, endpoint=False)
        else:
            if x.ndim != 1:
                raise ValueError(f"Array x (dim={x.ndim}), should be one-dimensional")
            if(x.size != y.size):
                raise ValueError(f"Array x size = {x.size} should be equal to array y size = {y.size}")
        return x

def _gen_op(op):
    return lambda self, other : self._op_helper(other, op)
for op in ["__" + op + "__" for op 
in ["add", "sub", "mul", "truediv", "iadd", "isub", "imul", "idiv"]]:
    setattr(Series, op, _gen_op(op))

def _gen_np_apply(fun):
    return lambda self: Series(getattr(np, fun)(self._y), self._x)
for fun in ["abs", "real", "imag", "angle"]:
    setattr(Series, fun, _gen_np_apply(fun))



import pyvisa
from typing import Union
from ..utils import check_type

class Wavemeter:

    def __init__(self, rm: pyvisa.ResourceManager, computer_name: str = "Wavemeter", port: int = 8001, **ignored):
        res_name = "::".join(("TCPIP", computer_name, str(port), "SOCKET"))
        self._res = rm.open_resource(res_name, write_termination='\n', read_termination='\n')

    @property
    def identity(self):
        return self._res.query("*IDN?")

    def _measure(self, channels: Union[int, tuple], what: str):
        if isinstance(channels, int):
            channels = (channels,)
        check_type((int), *channels)
        resp: str = self._res.query("meas:" + what + ' ' + ','.join([str(ch) for ch in channels]))
        vals = [float(v) for v in resp.split(",")]
        if len(vals) == 1:
            return vals[0]
        return vals

    def frequency(self, channels: Union[int, tuple]):
        return self._measure(channels, "freq")

    def wavelength(self, channels: Union[int, tuple]):
        return self._measure(channels, "wav")



import socket
import selectors
import types

class Server:

    def __init__(self) -> None:        
        self.read_termination = '\n'
        self.write_termination = '\n'
        self.port = 8001
        self.processor = lambda msg : msg        
        self._sel = selectors.DefaultSelector()
        self._buf = {}

    def accept_wrapper(self, sock):
        conn, addr = sock.accept()  # Should be ready to read
        id = addr[0] + ':' + str(addr[1])
        self._buf[id] = ""
        print("Accepted connection from", addr)
        conn.setblocking(False)
        data = types.SimpleNamespace(addr=addr, inb=b"", outb=b"")
        # events = selectors.EVENT_READ | selectors.EVENT_WRITE
        events = selectors.EVENT_READ
        self._sel.register(conn, events, data=data)


    def service_connection(self, key, mask):
        sock = key.fileobj
        data = key.data
        id = data.addr[0] + ':' + str(data.addr[1])
        if mask & selectors.EVENT_READ:
            try: recv_data = sock.recv(1024)  # Should be ready to read
            except ConnectionError:
                print("Connection with", id, "was aborted.")
                recv_data = None
            if recv_data:
                # print("Received:", recv_data)
                msgs = self.parse_data(recv_data, id)
                reps = self.process_data(msgs)
                for msg, rep in zip(msgs, reps):
                    print("From", id, "\n[Received]", msg, "\n[Replied ]", rep)
                for rep in reps:
                    if rep:
                        sock.sendall((rep + self.write_termination).encode())
            else:
                print("Closing connection to", id)
                del self._buf[id]
                self._sel.unregister(sock)
                sock.close()
        # if mask & selectors.EVENT_WRITE:
        #     if data.outb:
        #         print("echoing", repr(data.outb), "to", data.addr)
        #         sent = sock.send(data.outb)  # Should be ready to write
        #         data.outb = data.outb[sent:]

    def parse_data(self, data, id):
        self._buf[id] = self._buf[id] + data.decode()
        msgs = self._buf[id].split(self.read_termination)
        self._buf[id] = msgs[-1]
        return msgs[:-1]

    def process_data(self, msgs):
        return [self.processor(msg) for msg in msgs]

    def run(self):
        host = socket.gethostname()
        lsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        lsock.bind((host, self.port))
        lsock.listen()
        lsock.setblocking(False)
        self._sel.register(lsock, selectors.EVENT_READ, data=None)

        try:
            while True:
                print("Listening on", (host, self.port))
                events = self._sel.select()
                for key, mask in events:
                    if key.data is None:
                        self.accept_wrapper(key.fileobj)
                    else:
                        self.service_connection(key, mask)
        except KeyboardInterrupt:
            print("Caught keyboard interrupt, exiting...")
        finally:
            self._sel.close()

if (__name__ == "__main__"):
    serv = Server()
    serv.run()



import os
import importlib.metadata
# __version__ = importlib.metadata.version('labpython')
__version__ = '0.3.0'
dirname = os.path.dirname(__file__)
__doc__ = f"`labpython == {__version__}`\n"
try:
    with open(os.path.join(dirname, '../README.md')) as f:
        __doc__ += f.read()
except:
    pass



import jsbeautifier
import json

_unit_map = {
    'n': 1e-9, 'u': 1e-6, 'm': 1e-3, 'k': 1e3
}

def _split_unit(s:str):
    for i in range(len(s)):
        if s[i].isalpha():
            break
    return s[:i].strip(), s[i:].strip()

def str_to_value(s, base=1.):
    v, u = _split_unit(s)
    if len(u) == 1:
        mult = 1.
    else:
        mult = _unit_map[u[0]]
        mult /= base
    return float(v) * mult

def json_dumps_compact(data):
    opts = jsbeautifier.default_options()
    opts.indent_size = 2
    return jsbeautifier.beautify(json.dumps(data), opts)

def check_type(types, *vars):
    for v in vars:
        if not isinstance(v, types):
            raise TypeError(f"{type(v)} is not of type(s) {types}")

def floatify(v, precision=6):
    if isinstance(v, int):
        return (str(v) + '.0')
    if isinstance(v, float):
        if v.is_integer():
            return str(v)
        return (f"{v:.{precision}f}".rstrip('0'))
    raise TypeError(f"Variable of type {type(v)} is not a float" )

def intify(v):
    if isinstance(v, int):
        return str(v)
    raise TypeError(f"Variable of type {type(v)} is not an integer" )

def to_enum(v, Type):
    if isinstance(v, str):
        return Type[v.upper().replace(' ','_')]
    return Type(v)

def list_visa_devices(rm):
    res = [(str(inst.alias), str(inst.resource_name)) for inst in rm.list_resources_info().values()]
    res.insert(0, ("Alias", "Resource name"))
    for el in res:
        print(f"{el[0]:>15}  {el[1]}")

def in_bounds(value, bounds, margin=0., rel=0.) -> bool:
    '''Check if `value` is between `bounds` (pair of numbers).
    If `bounds = [l, r]` then values closer than
    `max(margin, (r - l) * rel)` to bounds are considered out of bounds.
    '''
    l, r = bounds
    if l >= r:
        return False
    off = max(margin, (r - l) * rel)
    l += off
    r -= off
    return l < value < r


try:
    import PyDAQmx as dmx
except Exception as e:
    print(e)
import numpy as np
from typing import Union

class DAQmx:

    def __init__(self, dev: str = "Dev1", channels: Union[str, tuple] = (),  freq=1000., time=0.01, trig: str = None, t0 = 0., **ignored):
        self._running = False
        self._dev = dev
        self._time = time - t0
        self._freq = freq
        self._samples = int(self._time * self._freq)
        self._pre_samples = 0
        self._triggered = False
        # SetSampQuantSampPerChan(uInt64) # to change later number of samples
        self._task = dmx.Task()
        if isinstance(channels, str):
            channels = (channels,)
        chs = ','.join([DAQmx._dev_path_join(dev, ch) for ch in channels])
        if not chs:
            raise ValueError("At least one channel must be specified")
        # CreateAIVoltageChan(physicalChannel: str, nameToAssignToChannel: str, terminalConfig: enum, minVal: float, maxVal: float, units: enum, None);
        self._task.CreateAIVoltageChan(chs, "", dmx.DAQmx_Val_Cfg_Default, -10., 10., dmx.DAQmx_Val_Volts, None)
        # CfgSampClkTiming(source: str, rate: float, activeEdge: enum, sampleMode: enum, sampsPerChan: int);
        self._task.CfgSampClkTiming("", self._freq, dmx.DAQmx_Val_Rising, dmx.DAQmx_Val_FiniteSamps, self._samples)
        if trig:
            self.set_trigger(trig, t0)
        self._chs_n = DAQmx._dmx_get(self._task.GetTaskNumChans, dmx.uInt32)

    def set_trigger(self, trig, t0):
        if self._triggered:
            raise ValueError("Trigger is already set up")
        self._triggered = True
        if t0 > 0:
            raise ValueError(f"t0 must be negative or zero (is {t0})")
        if t0 == 0:
            self._task.CfgDigEdgeStartTrig(DAQmx._dev_path_join(self._dev, trig), dmx.DAQmx_Val_Rising)
        else:
            self._pre_samples = int(max(2, -t0 * self._freq))
            self._task.CfgDigEdgeRefTrig(DAQmx._dev_path_join(self._dev, trig), dmx.DAQmx_Val_Rising, self._pre_samples)


    def start(self):
        if self._running:
            raise ValueError("Measurement already running. Must be read before starting again")
        self._running = True
        self._task.StartTask()

    def read(self, timeout=None):
        if not self._running:
            if self._triggered:
                raise ValueError("Measurement not running. Use start() before read()")
            else:
                self.start()
        if timeout is None:
            timeout = self.time + 10.
        data = np.zeros((self._chs_n, self._samples), dtype=np.float64)
        buf = data.ravel()
        # ReadAnalogF64(numSampsPerChan=int, timeout=float[sec], fillMode=enum, readArray=numpy.array, arraySizeInSamps=int, sampsPerChanRead=int_p, None);
        self._task.ReadAnalogF64(self._samples, float(timeout), dmx.DAQmx_Val_GroupByChannel, buf, buf.size, dmx.byref(dmx.int32()), None)
        self._task.StopTask()
        self._running = False
        return data

    def space(self):
        return np.linspace(self.t0, self.time + self.t0, self.samples, endpoint=False)

    @property
    def freq(self):
        return DAQmx._dmx_get(self._task.GetSampClkRate, dmx.float64)
    @freq.setter
    def freq(self, f):
        self._freq = f
        self._samples = int(self._time * self._freq)
        self._task.CfgSampClkTiming("", self._freq, dmx.DAQmx_Val_Rising, dmx.DAQmx_Val_FiniteSamps, self._samples)

    @property
    def t0(self):
        return -self._pre_samples / self.freq
    @t0.setter
    def t0(self, v):
        raise NotImplementedError

    @property
    def time(self):
        return self.samples / self.freq
        # return self._time
    @time.setter
    def time(self, t):
        self._time = t
        self._samples = int(self._time * self._freq)
        self._task.CfgSampClkTiming("", self._freq, dmx.DAQmx_Val_Rising, dmx.DAQmx_Val_FiniteSamps, self._samples)

    @property
    def samples(self):
        return self._samples

    @property
    def chs_n(self):
        return self._chs_n

    @staticmethod
    def _dev_path_join(*args):
        return '/' + '/'.join([arg.strip('/') for arg in args])

    @staticmethod
    def _dmx_get(getter, type):
        var = type()
        getter(dmx.byref(var))
        return var.value



import pyvisa
from ..utils import floatify, intify, str_to_value
from ..types import IndexedProperty

unit = 1e-6

class DmtCS:
    def __init__(self, rm: pyvisa.ResourceManager, dev='DMT', use_nimax_settings=False, **ignored):
        """Initialize connection to DMT current source. Fail if connection cannot be established.
        Parameters
        ----------
        rm
            VISA resource manager, e.g. returned by `pyvisa.ResourceManager()`
        dev: str = 'DMT'
            device name or alias, as returned by `labpy.utils.list_visa_devices` or visible in NI MAX
        use_nimax_settings: bool = False
            use connection settings saved in NI MAX
        Examples
        --------
        Check connection to DMT current source (change device name if necessary):
        ```python
        import pyvisa
        rm = pyvisa.ResourceManager()
        dmt = DmtCS(rm, 'DMT')
        print(dmt.identity)
        ```
        """
        if use_nimax_settings:
            kwargs = {'access_mode': 4}
        else:
            kwargs = {'baud_rate': 115200}
        self._res = rm.open_resource(dev, write_termination='\r\n', read_termination='\r\n', **kwargs)
        self.current = IndexedProperty(self.get_current, self.set_current)
        """array_like[float]:
        Array-like property for setting and getting current (uA) in respective channel (numbered from 1)

        Examples
        --------
        Set current to 50.5 uA in channel 2:
        ```python
        dmt.curent[2] = 50.5
        ```
        Get current to restore it later:
        ```python
        old_current = dmt.current[5]
        dmt.current[5] = 42
        ...
        dmt.current[5] = old_current
        ```
        """
        self.high_range = IndexedProperty(self.get_high_range, self.set_high_range)
        """array_like[{False: 4 mA, True: 40 mA}]:
        Array-like property for setting and getting current range in respective channel (numbered from 1)

        Examples
        --------
        Set current range to high (40 mA) in all channels
        ```python
        for i in range(1, 8+1):
            dmt.high_range[i] = True
        ```
        Notes
        -----
        Current range change will take effect only after setting current in the same channel:
        ```python
        dmt.high_range[5] = False
        # Nothing yet happens
        dmt.current[5] = 20.5
        # Range and current are changed simultanously
        ```
        """
        self._high_range_cache = {k: None for k in range(1, 8+1)}

    @property
    def identity(self):
        """str: Read only property returning identify string"""
        return self._res.query("!idn")

    def set_current(self, ch, v):
        if ch not in range(1, 8+1):
            raise ValueError("Channel index must be an integer from 1 to 8")
        if isinstance(v, (list, tuple)):
            v, high_range = v
            self.high_range[ch] = high_range
        rng = '40mA' if self.high_range[ch] else '4mA'
        message = ';'.join(['!set', str(ch), rng , floatify(v, 3)])
        self._res.write(message)

    def get_current(self, ch):
        if ch not in range(1, 8+1):
            raise ValueError("Channel index must be an integer from 1 to 8")
        resp = self._res.query("!chk;" + str(ch))
        return str_to_value(resp.split(';')[3], base=unit)
        
    def set_high_range(self, ch, v):
        if ch not in range(1, 8+1):
            raise ValueError("Channel index must be an integer from 1 to 8")
        self._high_range_cache[ch] = bool(v)

    def get_high_range(self, ch):
        if ch not in range(1, 8+1):
            raise ValueError("Channel index must be an integer from 1 to 8")
        if self._high_range_cache[ch] is None:
            rng = self._res.query("!chk;" + str(ch)).split(';')[2]
            if rng == '4mA':
                self._high_range_cache[ch] = False
            elif rng == '40mA':
                self._high_range_cache[ch] = True
            else:
                raise ValueError(f"Unknown range {rng}")
        return self._high_range_cache[ch]
        


import pyvisa

class TB3000AomDriver:
    def __init__(self, rm: pyvisa.ResourceManager, dev: str, use_nimax_settings = True, **ignored):
        if use_nimax_settings:
            kwargs = {'access_mode': 4}
        else:
            kwargs = {'baud_rate': 19200}
        self._res = rm.open_resource(dev, **kwargs, write_termination='\n', read_termination='\n')

    @property
    def identity(self):
        return self._res.query("*IDN?")

    @property
    def amplitude(self):
        return float(self._res.query('AOn?'))/100
    @amplitude.setter
    def amplitude(self, v):
        v = round(min(max(v, 10.), 100.) * 10)
        self._res.write("AOn " + str(v * 10))

    def output(self, set=True):
        if bool(set):
            self._res.write("OUT_on")
        else:
            self._res.write("OUT_off")


import pyvisa
from enum import Enum
from ..utils import floatify, intify, to_enum

class ArduinoPulseGen:

    class TimeUnit(Enum):
        CYCLE, US, MS, S = 0, 1, 2, 3

    def __init__(self, rm: pyvisa.ResourceManager, dev: str, useNiMaxSettings = True,
            portmap = {}, time_unit='ms', **ignored):
        access_mode = 4 if useNiMaxSettings else 0
        self._res = rm.open_resource(dev, access_mode=access_mode, write_termination='\n', read_termination='\n')
        self.reset_full()
        self.portmap = portmap
        self.time_unit = time_unit

    def _map_ch(self, ch):
        if isinstance(ch, str):
            ch = self.portmap[ch]
        return intify(ch)

    @property
    def identity(self):
        return self._res.query("*IDN?")

    def status(self, debug: bool = False):
        com = "puls? -l" if debug else "puls?"
        return self._res.query(com)

    @property
    def time_unit(self):
        return self._res.query("syst:unit?")
    @time_unit.setter
    def time_unit(self, v):
        unit_str = to_enum(v, self.TimeUnit).name.lower()
        self._res.write("syst:unit " + unit_str)

    def time_unit_store(self):
        self._res.write("syst:unit:store")
    # def time_unit_recall(self):
    #     self._res.write("syst:unit:rec")

    def set_ch(self, val, chs):
        if isinstance(chs, (int, str)):
            chs = (chs,)
        chs_str = ','.join([self._map_ch(ch) for ch in chs])
        cmd = "outp:on " if bool(val) else "outp:off "
        self._res.write(cmd + chs_str)

    def xon(self, chs):
        if isinstance(chs, (int, str)):
            chs = (chs,)
        chs_str = ','.join([self._map_ch(ch) for ch in chs])
        self._res.write("outp:xon " + chs_str)

    def on(self, chs):
        self.set_ch(True, chs)

    def off(self, chs = ()):
        self.set_ch(False, chs)

    def _add(self, chs, pulses, ver):
        if isinstance(pulses, (float, int)):
            pulses = (pulses,)
        if len(pulses) == 0:
            return
        pulses_str = ','.join([floatify(v) for v in pulses])
        if isinstance(chs, (str, int)):
            chs = (chs,)
        for ch in chs:
            self._res.write("puls:" + ver + ' ' + self._map_ch(ch) + ',' + pulses_str)
    def add(self, chs, pulses):
        self._add(chs, pulses, ver="add")

    def xadd(self, chs, pulses):
        self._add(chs, pulses, ver="xadd")

    def reset(self, ch=None):
        if ch:
            self._res.write("puls:reset " + self._map_ch(ch))
        else:
            self._res.write("puls:reset")

    def reset_full(self):
            self._res.write("*rst")
            # self._res.write("outp:off") #Quick fix, remove after updating Duo

    def run(self):
        self._res.write("puls:run")

if __name__ == "__main__":
    visa = pyvisa.ResourceManager()
    duo = ArduinoPulseGen(visa, "Arduino")
    # import constants
    # duo.portmap = constants.arduino.portmap
    print(duo.identity)
    duo.time_unit = "ms"
    print(duo.time_unit)
    # duo.reset()
    # duo.add("daqTrig", (0, 1.00341234))
    # duo.add("extra", (0, 2.1134312344))
    # duo._res.query("puls?")
    # duo.run()


class DataList(list):
    def __init__(self, init=[]):
        super().__init__(init)
        if hasattr(init, '__dict__'):
            for k, v in init.__dict__.items():
                self.__dict__[k] =v

    def __getitem__(self, idx):
        if isinstance(idx, str):
            return [el[idx] for el in super().__iter__()]
        elif isinstance(idx, (list, tuple)):
            # Explicit super() because of list comprehension scoping
            return [[el[id] for el in super(DataList, self).__iter__()] for id in idx]
        else:
            return super().__getitem__(idx)
    def __repr__(self) -> str:
        resp = 'Meta:\n' + self.meta.__repr__()
        if len(self) < 3:
            resp += '\nData:\n' +  super().__repr__()
        else:
            resp += '\nData:\n' + self[0].__repr__() + f'\n... {len(self) - 2} ...\n' + self[-1].__repr__()
        return resp
    @property
    def meta(self):
        return self.__dict__
    @property
    def data(self):
        return [el for el in super().__iter__()]

class IndexedProperty:
    def _default_getter(idx):
        raise NotImplementedError("Indexed property getter not implemmented")
    def _default_setter(idx, value):
        raise NotImplementedError("Indexed property setter not implemmented")
    def __init__(self, getter=None, setter=None):
        self._getitem = getter if getter is not None else IndexedProperty._default_getter
        self._setitem = setter if setter is not None else IndexedProperty._default_setter
    def __getitem__(self, idx):
        return self._getitem(idx)
    def __setitem__(self, idx, value):
        self._setitem(idx, value)

class NestedDict(dict):

    @staticmethod
    def _recursive_copy(dic):
        for k, v in dic.items():
            if isinstance(v, dict):
                dic[k] = v.copy()
                NestedDict._recursive_copy(dic[k])

    def _resolve_path(self, path, create=False):
        parent = self
        path = list(path)
        last = path.pop()
        for idx in path:
            if create and idx not in parent:
                parent[idx] = {}
            parent = parent[idx]
        return parent, last

    def __init__(self, dict={}):
        super().__init__(dict)
        NestedDict._recursive_copy(self)
        self.shadow = {}


    def copy(self):
        return NestedDict(self)

    def get(self, idx, dflt=None):
        try:
            return self[idx]
        except KeyError:
            return dflt

    def __getitem__(self, idx):
        if isinstance(idx, list):
            idx = tuple(idx)
        if isinstance(idx, tuple):
            if idx in self.shadow:
                return self.shadow[idx]
            else:
                node, idx = self._resolve_path(idx)
                return node[idx]
        else:
            return super().__getitem__(idx)

    def __setitem__(self, idx, value):
        if isinstance(idx, (list, tuple)):
            node, idx = self._resolve_path(idx, create=True)
            node[idx] = value
        else:
            return super().__setitem__(idx, value)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v

    def __repr__(self):
        if len(self.shadow) == 0:
            return super().__repr__()
        else:
            return super().__repr__() + '\nshadow: ' + self.shadow.__repr__()

class Average:

    def __init__(self, v = None):
        if v is not None:
            self.sum = v * 1.
        self.count = 0 if v is None else 1

    def add(self, v):
        if self.count == 0:
            self.sum = v * 1.
        else:
            self.sum += v
        self.count += 1

    @property
    def value(self):
        if self.count == 0:
            return 0.
        return self.sum / self.count

from .series import Series
